<?php

/**
 * @file
 * Contains \Drupal\smartling\Plugin\Action\TranslateNode.
 */

namespace Drupal\smartling\Plugin\Action;

use Drupal\Core\Action\ConfigurableActionBase;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Queue\QueueInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\smartling\Entity\SmartlingSubmission;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Translate entity.
 *
 * @Action(
 *   id = "smartling_translate_node_action",
 *   label = @Translation("Translate node to all configured languages"),
 *   type = "node"
 * )
 */
class TranslateNode extends ConfigurableActionBase implements ContainerFactoryPluginInterface {

  /**
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * The submission storage.
   *
   * @var \Drupal\smartling\SubmissionStorageInterface
   */
  protected $entityStorage;

  /**
   * The upload queue.
   *
   * @var \Drupal\Core\Queue\QueueInterface
   */
  protected $queue;

  /**
   * Creates TranslateNode action.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Queue\QueueInterface $queue
   *   The upload queue.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   Language manager service.
   * @param \Drupal\Core\Config\ImmutableConfig $config
   *   Immutable config instance that contains smartling settings.
   * @param \Drupal\Core\Entity\EntityStorageInterface $entity_storage
   *   SmartlingSubmission entity storage.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, QueueInterface $queue, LanguageManagerInterface $language_manager, ImmutableConfig $config, EntityStorageInterface $entity_storage) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->queue = $queue;
    $this->languageManager = $language_manager;
    $this->config = $config;
    $this->entityStorage = $entity_storage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static($configuration, $plugin_id, $plugin_definition,
      $container->get('queue')->get('smartling_upload'),
      $container->get('language_manager'),
      $container->get('config.factory')->get('smartling.settings'),
      $container->get('entity.manager')->getStorage('smartling_submission')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return ['locales' => []];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $languages = $this->languageManager->getLanguages();
    unset($languages[$this->languageManager->getCurrentLanguage()->getId()]);

    $config = $this->config->get('account_info.enabled_languages');
    $locales = [];
    foreach ($config as $language_code) {
      $locales[$language_code] = $languages[$language_code]->getName();
    }

    $form['locales'] = array(
      '#type' => 'checkboxes',
      '#title' => t('Select locales'),
      '#options' => $locales,
      '#default_value' => $this->configuration['locales'],
      '#required' => TRUE,
    );
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['locales'] = $form_state->getValue('locales');
  }

  /**
   * {@inheritdoc}
   */
  public function execute(ContentEntityInterface $entity = NULL) {
    $locales = $this->configuration['locales'];
    if (!$locales) {
      // Fallback to enabled languages.
      $config = $this->config->get('account_info.enabled_languages');
      $locales = [];
      foreach ($config as $language_code) {
        $locales[] = $language_code;
      }
    }

    // Make sure all submissions exists.
    foreach ($locales as $language_code) {
      $submission = $this->entityStorage->loadByProperties([
        'entity_id' => $entity->id(),
        'entity_entity' => $entity->getEntityTypeId(),
        'target_language' => $language_code,
      ]);
      if (!$submission) {
        $submission = SmartlingSubmission::getFromDrupalEntity($entity, $language_code);
        $submission->save();
      }
    }
    if (isset($submission)) {
      $item_data = [
        'entity_type' => $entity->getEntityTypeId(),
        'entity_id' => $entity->id(),
        'file_name' => $submission->getFileName(),
        'locales' => $locales,
      ];
      $this->queue->createItem($item_data);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, AccountInterface $account = NULL, $return_as_object = FALSE) {
    /** @var \Drupal\node\NodeInterface $object */
    // @todo check smartling permissions.
    $result = $object->access('update', $account, TRUE)
      ->andIf($object->status->access('edit', $account, TRUE));

    return $return_as_object ? $result : $result->isAllowed();
  }

}
