<?php

/**
 * @file
 * Contains \Drupal\smartling\Plugin\smartling\Source\ContentEntitySource.
 */

namespace Drupal\smartling\Plugin\smartling\Source;

use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemInterface;
use Drupal\Core\File\FileSystem;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\TypedData\OptionsProviderInterface;
use Drupal\Core\TypedData\Type\StringInterface;
use Drupal\Core\TypedData\PrimitiveInterface;
use Drupal\smartling\ApiWrapper\ApiWrapperInterface;
use Drupal\smartling\ApiWrapper\SmartlingApiWrapper;
use Drupal\smartling\SmartlingSubmissionInterface;
use Drupal\smartling\SourcePluginBase;
use Drupal\Core\Render\Element;
use Exception;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Serializer;

/**
 * Content entity source plugin controller.
 *
 * @SourcePlugin(
 *   id = "content",
 *   label = @Translation("Content Entity"),
 *   description = @Translation("Source handler for entities.")
 * )
 */
class ContentEntitySource extends SourcePluginBase implements ContainerFactoryPluginInterface {

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * @var \Symfony\Component\Serializer\Serializer
   */
  protected $serializer;

  /**
   * The smartling API wrapper.
   *
   * @var \Drupal\smartling\ApiWrapper\ApiWrapperInterface
   */
  protected $api;

  /**
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $entityStorage;

  /**
   * @var \Drupal\Core\File\FileSystem
   */
  protected $fileSystem;

  /**
   * @var \Drupal\smartling\SourcePluginInterface
   */
  protected $sourcePlugin;

  /**
   * Creates TranslateNode action.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   * @param \Symfony\Component\Serializer\Serializer $serializer
   * @param \Drupal\smartling\ApiWrapper\ApiWrapperInterface $smartling_api_wrapper
   *   The smartling API wrapper.
   * @param \Drupal\Core\Entity\EntityStorageInterface $entity_storage
   *   SmartlingSubmission entity storage.
   * @param \Drupal\Core\File\FileSystem $file_system
   * @param \Drupal\Core\Config\ImmutableConfig $config
   *   Immutable config instance that contains smartling settings.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, LoggerChannelInterface $logger, Serializer $serializer, ApiWrapperInterface $smartling_api_wrapper, EntityStorageInterface $entity_storage, FileSystem $file_system, ImmutableConfig $config) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->logger = $logger;
    $this->serializer = $serializer;
    $this->api = $smartling_api_wrapper;
    $this->entityStorage = $entity_storage;
    $this->fileSystem = $file_system;
    $this->config = $config;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.channel.smartling'),
      $container->get('serializer'),
      $container->get('smartling.api_wrapper'),
      $container->get('entity.manager')->getStorage('smartling_submission'),
      $container->get('file_system'),
      $container->get('config.factory')->get('smartling.settings')
    );
  }

  /**
   * Returns the data from the fields as a structure that can be processed by
   * the Smartling.
   *
   * @param \Drupal\smartling\SmartlingSubmissionInterface $smartling_item
   *   SmartlingData entity.
   *
   * @return array
   *
   * @throws \Exception
   */
  public function getData(SmartlingSubmissionInterface $smartling_item) {
    $entity = $smartling_item->getRelatedEntity();
    if (!$entity) {
      // @todo provide own exceptions.
      throw new Exception(t('Unable to load entity %type with id %id', [
        '%type' => $smartling_item->get('entity_type')->value,
        '%id' => $smartling_item->get('entity_id')->value,
      ]));
    }
    // @todo inject through the DI.
    $languages = \Drupal::languageManager()->getLanguages();
    $id = $entity->language()->getId();
    if (!isset($languages[$id])) {
      throw new Exception(t('Entity %entity could not be translated because the language %language is not applicable', [
        '%entity' => $entity->id(),
        '%language' => $entity->language()->getName(),
      ]));
    }

    if (!$entity->hasTranslation($smartling_item->get('original_language')->value)) {
      throw new Exception(t('The entity %id with translation %lang does not exist.', [
        '%id' => $entity->id(),
        '%lang' => $smartling_item->language()->getId(),
      ]));
    }

    $translation = $entity->getTranslation($smartling_item->get('original_language')->value);
    $data = $this->extractTranslatableData($translation);
    return $data;
  }

  /**
   * Extracts translatable data from an entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to get the the translatable data from.
   *
   * @return array $data
   *   Translatable data.
   */
  public function extractTranslatableData(ContentEntityInterface $entity) {

    // @todo Expand this list or find a better solution to exclude fields like
    //   content_translation_source.

    $field_definitions = $entity->getFieldDefinitions();
    $exclude_field_types = ['language'];
    $translatable_fields = array_filter($field_definitions, function (FieldDefinitionInterface $field_definition) use ($exclude_field_types) {
        return $field_definition->isTranslatable() && !in_array($field_definition->getType(), $exclude_field_types);
    });

    $data = array();
    /* @var FieldDefinitionInterface $field_definition */
    foreach ($translatable_fields as $key => $field_definition) {
      $field = $entity->get($key);
      foreach ($field as $index => $field_item) {
        $format = NULL;
        /* @var FieldItemInterface $field_item */
        foreach ($field_item->getProperties() as $property_key => $property) {
          // Ignore computed values.
          $property_definition = $property->getDataDefinition();
          // Ignore values that are not primitives.
          if (!($property instanceof PrimitiveInterface)) {
            continue;
          }
          $translate = TRUE;
          // Ignore properties with limited allowed values or if they're not strings.
          if ($property instanceof OptionsProviderInterface || !($property instanceof StringInterface)) {
            $translate = FALSE;
          }
          // All the labels are here, to make sure we don't have empty labels in
          // the UI because of no data.
          if ($translate == TRUE) {
            $data[$key]['#label'] = $field_definition->getName();
            $data[$key][$index]['#label'] = t('Delta #@delta', array('@delta' => $index));
          }
          $data[$key][$index][$property_key] = array(
            '#label' => $property_definition->getLabel(),
            '#text' => $property->getValue(),
            '#translate' => $translate,
          );

          if ($property_definition->getDataType() == 'filter_format') {
            $format = $property->getValue();
          }
        }
        // Add the format to the translatable properties.
        if (!empty($format)) {
          foreach ($data[$key][$index] as $name => $value) {
            if (isset($value['#translate']) && $value['#translate'] == TRUE) {
              $data[$key][$index][$name]['#format'] = $format;
            }
          }
        }
      }
    }

    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function saveTranslation(SmartlingSubmissionInterface $smartling_item) {
    /* @var \Drupal\Core\Entity\ContentEntityInterface $entity */
    $entity = $smartling_item->getRelatedEntity();
    $file = $smartling_item->getFileName();
    // @todo process xml file and retrieve data.
    $xml = [];
    $this->doSaveTranslations($entity, $data, $smartling_item->get('target_language'));
  }

  /**
   * {@inheritdoc}
   */
  public function getItemTypes() {
    $entity_types = \Drupal::entityManager()->getDefinitions();
    $types = array();
    $content_translation_manager = \Drupal::service('content_translation.manager');
    foreach ($entity_types as $entity_type_name => $entity_type) {
      if ($content_translation_manager->isEnabled($entity_type->id())) {
        $types[$entity_type_name] = $entity_type->getLabel();
      }
    }
    return $types;
  }

  /**
   * {@inheritdoc}
   */
  public function getSourceLangCode(SmartlingSubmissionInterface $smartling_item) {
    $entity = $smartling_item->getRelatedEntity();
    return $entity->getUntranslated()->language()->getId();
  }

  /**
   * Upload content from give smartling entity to smartling as file.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The content entity object to serialize.
   * @param string $file_name
   *   The unique file name to identify submission.
   * @param array $locales
   *   List of locales ask Smartling translate source entity to.
   *
   * @return string|NULL
   *   The upload status constant or NULL when file can
   */
  public function uploadSubmissions(ContentEntityInterface $entity, $file_name, array $locales) {
    $xml = $this->serializer->serialize($entity, 'smartling_xml');

    $dir = 'public://smartling/content';
    $success = file_prepare_directory($dir, FILE_CREATE_DIRECTORY | FILE_MODIFY_PERMISSIONS);
    $success = $success && file_save_htaccess($dir);
    if (!$success) {
      return;
    }

    $file_path = file_unmanaged_save_data($xml, $dir . '/' . $file_name, FILE_EXISTS_REPLACE);
    if (!$file_path) {
      return;
    }

    return $this->api->uploadFile(
      $this->fileSystem->realpath($file_path),
      $file_name,
      ApiWrapperInterface::TYPE_XML, $locales
    );
  }

  /**
   * Download latest version of translation for the Smartling submission.
   *
   * @param \Drupal\smartling\SmartlingSubmissionInterface $submission
   *   Smartling entity object.
   *
   * @return bool
   *   TRUE when the submission downloaded and translation created or updated.
   */
  public function downloadSubmission(SmartlingSubmissionInterface $submission) {
    $result = $this->api->downloadFile($submission);
    if ($result) {
      // Apply translation to entity fields.
      $entity = $submission->getRelatedEntity();
      $entity_type = $entity->getEntityType();
      $context = ['entity_type' => $entity->getEntityTypeId()];
      // Provide bundle value to properly create entity.
      if ($entity_type->hasKey('bundle')) {
        $context['bundle_key'] = $entity_type->getKey('bundle');
        $context['bundle_value'] = $entity->bundle();
      }
      $entity_class = $entity_type->getClass();
      /** @var \Drupal\Core\Entity\ContentEntityInterface $new_entity */
      $new_entity = $this->serializer->deserialize($result, $entity_class, 'smartling_xml', $context);

      $lang_code = $submission->get('target_language')->value;
      // Switch entity to current language.
      if ($entity->hasTranslation($lang_code)) {
        $translation = $entity->getTranslation($lang_code);
      }
      else {
        $translation = $entity->addTranslation($lang_code);
      }
      foreach ($new_entity->_restSubmittedFields as $field) {
        if (isset($context['bundle_key']) && $field === $context['bundle_key']) {
          // Skip changing bundle value.
          continue;
        }
        $field_items = $translation->get($field);
        if ($field_items->getFieldDefinition()->isTranslatable()) {
          // Update only translatable fields.
          $field_items->setValue($new_entity->get($field)->getValue());
        }
      }
      $translation->save();
      // @todo Save file name and properly update submission.
      $submission
        ->set('progress', 100)
        ->save();
      return TRUE;
    }
    return FALSE;
  }

  /**
   * @param $data
   * @param $file_name
   *
   * @return \Drupal\file\FileInterface
   *
   * @todo throw exception if something is bad. Also add relation to smartling
   * entity.
   */
  public function saveFile($data, $file_name) {
    return file_save_data($data, 'private://' . $file_name);
  }

  /**
   * Saves translation data in an entity translation.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity for which the translation should be saved.
   * @param \Drupal\Core\Entity\ContentEntityInterface $translated_entity
   *   The translation data for the fields.
   * @param string $target_langcode
   *   The target language.
   */
  protected function doSaveTranslations(ContentEntityInterface $entity, ContentEntityInterface $translated_entity, $target_langcode) {
    // If the translation for this language does not exist yet, initialize it.
    if (!$entity->hasTranslation($target_langcode)) {
      $entity->addTranslation($target_langcode, $entity->toArray());
    }

    $translation = $entity->getTranslation($target_langcode);

    foreach ($translation as $name => $field_data) {
      foreach (Element::children($field_data) as $delta) {
        $field_item = $field_data[$delta];
        foreach (Element::children($field_item) as $property) {
          $property_data = $field_item[$property];
          // If there is translation data for the field property, save it.
          if (isset($property_data['#translation']['#text'])) {
            $translation->get($name)
              ->offsetGet($delta)
              ->set($property, $property_data['#translation']['#text']);
          }
        }
      }
    }
    $translation->save();
  }

}
